import json
import os
import typing

import tornado.web

from tornado_swagger._builders import generate_doc_from_endpoints
from tornado_swagger._handlers import SwaggerHomeHandler

STATIC_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), 'swagger_ui'))


def export_swagger(routes: typing.List[tornado.web.URLSpec],
                   *,
                   api_base_url: str = '/',
                   description: str = 'Swagger API definition',
                   api_version: str = '1.0.0',
                   title: str = 'Swagger API',
                   contact: str = '',
                   schemes: list = None,
                   security_definitions: dict = None
                   ):
    return generate_doc_from_endpoints(
        routes,
        api_base_url=api_base_url,
        description=description,
        api_version=api_version,
        title=title,
        contact=contact,
        schemes=schemes,
        security_definitions=security_definitions,
    )


def setup_swagger(routes: typing.List[tornado.web.URLSpec],
                  *,
                  swagger_url: str = '/api/doc',
                  api_base_url: str = '/',
                  description: str = 'Swagger API definition',
                  api_version: str = '1.0.0',
                  title: str = 'Swagger API',
                  contact: str = '',
                  schemes: list = None,
                  security_definitions: dict = None
                  ):
    swagger_schema = generate_doc_from_endpoints(
        routes,
        api_base_url=api_base_url,
        description=description,
        api_version=api_version,
        title=title,
        contact=contact,
        schemes=schemes,
        security_definitions=security_definitions,
    )

    _swagger_url = ('/{}'.format(swagger_url)
                    if not swagger_url.startswith('/')
                    else swagger_url)
    _base_swagger_url = _swagger_url.rstrip('/')

    routes += [
        tornado.web.url(_swagger_url, SwaggerHomeHandler),
        tornado.web.url('{}/'.format(_base_swagger_url), SwaggerHomeHandler),
    ]

    with open(os.path.join(STATIC_PATH, 'ui.html'), 'r') as f:
        SwaggerHomeHandler.SWAGGER_HOME_TEMPLATE = (
            f.read().replace(
                '{{ SWAGGER_SCHEMA }}',
                json.dumps(swagger_schema)
            )
        )



import collections
import inspect
import os
import re
import typing

import tornado.web
import yaml

SWAGGER_TEMPLATE = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'templates', 'swagger.yaml')
)
SWAGGER_DOC_SEPARATOR = '---'


def _extract_swagger_definition(endpoint_doc):
    endpoint_doc = endpoint_doc.splitlines()

    # Find Swagger start point in doc
    for i, doc_line in enumerate(endpoint_doc):
        if SWAGGER_DOC_SEPARATOR in doc_line:
            end_point_swagger_start = i + 1
            endpoint_doc = endpoint_doc[end_point_swagger_start:]
            break
    return '\n'.join(endpoint_doc)


def extract_swagger_docs(endpoint_doc):
    endpoint_doc = _extract_swagger_definition(endpoint_doc)

    # Build JSON YAML Obj
    try:
        end_point_swagger_doc = yaml.safe_load(endpoint_doc)
        if not isinstance(end_point_swagger_doc, dict):
            raise yaml.YAMLError()
    except yaml.YAMLError:
        end_point_swagger_doc = {
            'description': 'Swagger document could not be loaded from docstring',
            'tags': ['Invalid Swagger']
        }
    return end_point_swagger_doc


def _try_extract_doc(func):
    return inspect.unwrap(func).__doc__


def _build_doc_from_func_doc(handler):
    out = {}

    for method in handler.SUPPORTED_METHODS:
        method = method.lower()
        doc = _try_extract_doc(getattr(handler, method))

        if doc is not None and '---' in doc:
            out.update({
                method: extract_swagger_docs(doc)
            })

    return out


def _try_extract_args(method_handler):
    return inspect.getfullargspec(inspect.unwrap(method_handler)).args[1:]


def _extract_parameters_names(handler, parameters_count):
    if parameters_count == 0:
        return []

    parameters = ['{?}' for _ in range(parameters_count)]

    for method in handler.SUPPORTED_METHODS:
        method_handler = getattr(handler, method.lower())
        args = _try_extract_args(method_handler)

        if len(args) > 0:
            for i, arg in enumerate(args):
                if set(arg) != {'_'}:
                    parameters[i] = arg

    return parameters


def _format_handler_path(route):
    brackets_regex = re.compile(r'\(.*?\)')
    parameters = _extract_parameters_names(route.target, route.regex.groups)
    route_pattern = route.regex.pattern

    for i, entity in enumerate(brackets_regex.findall(route_pattern)):
        route_pattern = route_pattern.replace(entity, '{%s}' % parameters[i], 1)

    return route_pattern[:-1]


def nesteddict2yaml(d, indent=10, result=''):
    for key, value in d.items():
        result += ' ' * indent + str(key) + ':'
        if isinstance(value, dict):
            result = nesteddict2yaml(value, indent + 2, result + '\n')
        else:
            result += ' ' + str(value) + '\n'
    return result


def generate_doc_from_endpoints(routes: typing.List[tornado.web.URLSpec],
                                *,
                                api_base_url,
                                description,
                                api_version,
                                title,
                                contact,
                                schemes,
                                security_definitions):
    from tornado_swagger.model import swagger_models
    # Clean description
    _start_desc = 0
    for i, word in enumerate(description):
        if word != '\n':
            _start_desc = i
            break
    cleaned_description = '    '.join(description[_start_desc:].splitlines())

    # The Swagger OBJ
    swagger = {
        'swagger': '2.0',
        'info': {
            'description': cleaned_description,
            'version': api_version,
            'title': title
        },
        'basePath': api_base_url,
    }
    if contact:
        swagger['info']['contact'] = {
            'name': contact
        }
    if security_definitions:
        swagger['securityDefinitions'] = nesteddict2yaml(security_definitions)

    swagger['schemes'] = schemes
    swagger['paths'] = collections.defaultdict(dict)
    swagger['definitions'] = swagger_models

    for route in routes:
        swagger['paths'][_format_handler_path(route)].update(_build_doc_from_func_doc(route.target))

    return swagger



import tornado.web


class TornadoHandler(tornado.web.RequestHandler):
    def data_received(self, chunk: bytes):
        pass


class SwaggerHomeHandler(TornadoHandler):
    SWAGGER_HOME_TEMPLATE = ''

    def get(self):
        self.write(self.SWAGGER_HOME_TEMPLATE)



from tornado_swagger._builders import extract_swagger_docs

swagger_models = {}


def _save_model_doc(model):
    global swagger_models
    doc = model.__doc__

    if doc is not None and '---' in doc:
        swagger_models[model.__name__] = extract_swagger_docs(doc)


def register_swagger_model(model):
    _save_model_doc(model)
    return model



from os.path import dirname
from os.path import join

from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test as TestCommand

from tornado_swagger import __version__

PACKAGE_NAME = 'tornado-swagger'
DESCRIPTION = 'Swagger API Documentation builder for tornado server'
HOME_URL = 'https://github.com/mrk-andreev/tornado-swagger'
DOWNLOAD_URL = 'https://pypi.org/project/tornado-swagger/#files'
MAINTAINER = 'Mark Andreev'
MAINTAINER_EMAIL = 'mark.andreev@gmail.com'
LICENSE = 'MIT License'
CLASSIFIERS = [
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

with open(join(dirname(__file__), 'requirements.txt')) as f:
    PACKAGES_REQUIRED = f.read().splitlines()

with open(join(dirname(__file__), 'README.md')) as f:
    LONG_DESCRIPTION = f.read()


class PyTest(TestCommand):
    user_options = []

    def run(self):
        import subprocess
        import sys
        errno = subprocess.call([sys.executable, '-m', 'pytest', 'tests'])
        raise SystemExit(errno)


def setup_package():
    setup(
        name=PACKAGE_NAME,
        version=__version__,
        install_requires=PACKAGES_REQUIRED,
        url=HOME_URL,
        download_url=DOWNLOAD_URL,
        license=LICENSE,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        packages=find_packages(),
        include_package_data=True,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        classifiers=CLASSIFIERS,
        tests_require=['pytest'],
        cmdclass=dict(test=PyTest)
    )


if __name__ == "__main__":
    setup_package()
